/* ************************************************************************** */
/*                                       _____           _                    */
/*  log.c                               |_   _|__  _   _| |__   ___ _   _     */
/*  | Project : libp7                     | |/ _ \| | | | '_ \ / _ \ | | |    */
/*                                        | | (_) | |_| | | | |  __/ |_| |    */
/*  By: thomas <thomas@touhey.fr>         |_|\___/ \__,_|_| |_|\___|\__, |.fr */
/*  Last updated: 2016/08/19 12:04:33                               |___/     */
/*                                                                            */
/* ************************************************************************** */

#include <stdio.h>
#include <ctype.h>
#include <libp7/internals.h>
#define min(A, B) ((A) < (B) ? (A) : (B))

/**
 *	log_mem_oct:
 *	Prints the octal interpretation of a max of two octets.
 *
 *	@arg	m		the memory zone to print
 *	@arg	n		the size of the memory zone
 */

static void log_mem_oct(unsigned char *m, size_t n)
{
	size_t l = 0;
	while (l < 16) {
		if (n)
			fprintf(stderr, "%02X", *m++);
		else
			fprintf(stderr, "  ");
		n -= !!n;

		if (l++ % 2)
			fprintf(stderr, " ");
	}
}

/**
 *	log_mem_asc:
 *	Prints the ascii interpretation of a max of two octets.
 *
 *	@arg	m		the memory zone to print
 *	@arg	n		the size of the memory zone
 */

static void log_mem_asc(unsigned char *m, size_t n)
{
	size_t l = 0;
	while (n-- && l++ < 16) {
		if (isprint(*m++))
			fprintf(stderr, "%c", *((char*)m - 1));
		else
			fprintf(stderr, ".");
	}
}

/**
 *	p7_log_mem:
 *	Print memory zone.
 *
 *	@arg	prefx	the line prefix
 *	@arg	m		the memory zone to print
 *	@arg	n		the size of the memory zone
 */

void p7_log_mem(const char *prefx, void *m, size_t n)
{
	unsigned char *p = m;
	if (!n) fprintf(stderr, "%s(nothing)\n", prefx);
	while (n > 0) {
		fprintf(stderr, "%s", prefx);
		log_mem_oct(p, n);
		log_mem_asc(p, n);
		fprintf(stderr, "\n");
		p += 16;
		n -= min(16, n);
	}
}

/** 
 *	p7_getcmdstring:
 *	Get command code string (useful for logging).
 *
 *	@arg	code		command code
 *	@return				the string
 */

const char *p7_getcmdstring(p7_commandcode_t code)
{
	static const char *sys_commands[] = {
		"restart the equipment",
		"get device info",
		"set link settings"
	}, *fls_commands[] = {
		"create directory on flash",
		"delete directory on flash",
		"rename directory on flash",
		"change working directory on flash",
		"file transfer request on flash",
		"file transfer on flash",
		"delete file on flash",
		"rename file on flash",
		"copy file on flash"
	};

	if (code >= 0x40) return fls_commands[code - 0x40];
	/* TODO: add MCS commands */
	else return sys_commands[code];
}

/**
 *	p7_geterrstring:
 *	Get error string (useful for logging).
 *
 *	@arg	code		error code
 *	@return				the string
 */

const char *p7_geterrstring(p7_errcode_t code)
{
	static const char *strings[] = {
		"should resend",
		"file exists, overwrite request",
		"will not overwrite",
		"overwrite impossible",
		"memory is full"
	};

	return (strings[code - 1]);
}

/**
 *	p7_gettermstring:
 *	Get termination cause string (useful for logging).
 *
 *	@arg	code		cause code
 *	@return				the string
 */

const char *p7_gettermstring(p7_termtype_t code)
{
	static const char *strings[] = {
		"default",
		"user requested",
		"timeouts",
		"on overwrite request"
	};

	return (strings[code]);
}

/**
 *	p7_getowstring:
 *	Get overwrite string (useful for logging).
 *
 *	@arg	code		overwrite code
 *	@return				the string
 */

const char *p7_getowstring(p7_commandoverwrite_t code)
{
	static const char *moo[] = {
		"request confirmation before overwriting",
		"terminate if file exists",
		"force overwrite"
	};

	return (moo[code]);
}
