/* ************************************************************************** */
/*                                       _____           _                    */
/*  packet/recv.c                       |_   _|__  _   _| |__   ___ _   _     */
/*  | Project : libp7                     | |/ _ \| | | | '_ \ / _ \ | | |    */
/*                                        | | (_) | |_| | | | |  __/ |_| |    */
/*  By: thomas <thomas@touhey.fr>         |_|\___/ \__,_|_| |_|\___|\__, |.fr */
/*  Last updated: 2016/08/16 12:44:08                               |___/     */
/*                                                                            */
/* ************************************************************************** */

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <libp7/internals.h>

/* ************************************************************************** */
/*  Data layouts                                                              */
/* ************************************************************************** */
/**
 *	packetdata_ackext_t:
 *	The extended ack packet data field layout.
 */

typedef struct {
	/* hardware identifier - ASCII */
	char hwid[8];
	/* processor identifier - ASCII */
	char cpuid[16];

	/* preprogrammed ROM capacity - ASCII-hex (Ko) */
	char preprog_rom_capacity[8];
	/* flash ROM capacity - ASCII-hex (Ko) */
	char flash_rom_capacity[8];
	/* RAM capacity - ASCII-hex (Ko) */
	char ram_capacity[8];

	/* preprogrammed ROM version - "xx.xx.xx" + type */
	char preprog_rom_version[16];

	/* bootcode version - "xx.xx.xx" + type */
	char bootcode_version[16];
	/* bootcode offset - ASCII-hex */
	char bootcode_offset[8];
	/* bootcode size - ASCII-hex (Ko) */
	char bootcode_size[8];

	/* OS version - "xx.xx.xx" + type */
	char os_version[16];
	/* OS offset - ASCII-hex */
	char os_offset[8];
	/* OS size - ASCII-hex (Ko) */
	char os_size[8];

	/* protocol version - "xx.x" */
	char protocol_version[4];
	/* product ID */
	char product_id[16];
	/* name set by user in SYSTEM */
	char username[16];
} packetdata_ackext_t;

/**
 *	packetdata_data_t:
 *	The data packet data field layout.
 */

typedef struct {
	char total_number[4];
	char current_number[4];
	unsigned char data[];
} packetdata_data_t;

/* ************************************************************************** */
/*  Useful functions                                                          */
/* ************************************************************************** */
/**
 *	p7_recv_getver:
 *	Get version from 16-bytes char buffer.
 *
 *	data is not a string as it doesn't finish with an '\0'.
 *
 *	@arg	ver			the version destination
 *	@arg	data		the raw version data ("xx.xx.xx" + type)
 */

static void p7_recv_getver(p7_version_t *ver, char data[16])
{
	char buf[9] = {0};

	/* get major */
	memcpy(buf, data, 2);
	ver->major = (int)strtoul(buf, NULL, 16);
	/* get minor */
	memcpy(buf, &data[3], 2);
	ver->minor = (int)strtoul(buf, NULL, 16);
	/* get rev */
	memcpy(buf, &data[6], 2);
	ver->rev = (int)strtoul(buf, NULL, 16);
	/* get type */
	memcpy(&ver->type, &data[8], 8);
	ver->type[8] = '\0';

	/* log */
	log_info("major: %#02x, minor: %#02x, rev: %#02x",
		ver->major, ver->minor, ver->rev);
}

/**
 *	p7_recv_getack:
 *	Get data from ack data field.
 *
 *	Layout is described in the fxReverse projet documentation.
 *
 *	@arg	ack			the ack packet destination
 *	@arg	data		the raw data
 */

static void p7_recv_getack(p7_packetack_t *ack, void *data)
{
	packetdata_ackext_t *d = data;
	char buf[9] = {0};
	char *t; /* temporary char pointer */

	/* log */
	log_info("ack packet is extended");

	/* hardware identifier */
	memcpy(ack->hwid, d->hwid, 8); ack->hwid[8] = '\0';
	log_info("hardware identifier is '%s'", ack->hwid);
	/* processor identifier */
	memcpy(ack->cpuid, d->cpuid, 16); ack->cpuid[16] = '\0';
	log_info("cpu identifier is '%s'", ack->cpuid);

	/* preprogrammed ROM capacity */
	memcpy(buf, d->preprog_rom_capacity, 8);
	ack->preprog_rom_capacity = (unsigned int)strtoul(buf, NULL, 16) * 1000;
	log_info("preprogrammed ROM capacity is %uo",
		ack->preprog_rom_capacity);
	/* flash ROM capacity */
	memcpy(buf, d->flash_rom_capacity, 8);
	ack->flash_rom_capacity = (unsigned int)strtoul(buf, NULL, 16) * 1000;
	log_info("flash ROM capacity is %uo",
		ack->flash_rom_capacity);
	/* RAM capacity */
	memcpy(buf, d->ram_capacity, 8);
	ack->ram_capacity = (unsigned int)strtoul(buf, NULL, 16) * 1000;
	log_info("RAM capacity is %uo", ack->ram_capacity);

	/* preprogrammed ROM version */
	log_info("preprogrammed ROM version is the following :");
	p7_recv_getver(&ack->preprog_rom_version, d->preprog_rom_version);

	/* bootcode version */
	log_info("bootcode version is the following :");
	p7_recv_getver(&ack->bootcode_version, d->bootcode_version);
	/* bootcode offset */
	memcpy(buf, d->bootcode_offset, 8); buf[8] = '\0';
	ack->bootcode_offset = (unsigned int)strtoul(buf, NULL, 16);
	log_info("bootcode offset is %u", ack->bootcode_offset);
	/* bootcode size */
	memcpy(buf, d->bootcode_size, 8); buf[8] = '\0';
	ack->bootcode_size = (unsigned int)strtoul(buf, NULL, 16);
	log_info("bootcode size is %u", ack->bootcode_size);

	/* OS version */
	log_info("OS version is the following :");
	p7_recv_getver(&ack->os_version, d->os_version);
	/* OS offset */
	memcpy(buf, d->os_offset, 8); buf[8] = '\0';
	ack->os_offset = (unsigned int)strtoul(buf, NULL, 16);
	log_info("OS offset is %#08x", ack->os_offset);
	/* OS size */
	memcpy(buf, d->os_size, 8); buf[8] = '\0';
	ack->os_size = (unsigned int)strtoul(buf, NULL, 16);
	log_info("OS size is %uo", ack->os_size);

	/* protocol version */
	memcpy(buf, d->protocol_version, 4); buf[1] = '\0', buf[4] = '\0';
	ack->protocol_version.major = (int)strtoul(buf, NULL, 16);
	ack->protocol_version.minor = (int)strtoul(&buf[2], NULL, '\0');
	log_info("protocol version is %d.%d",
		ack->protocol_version.major, ack->protocol_version.minor);
	/* product ID */
	memcpy(ack->product_id, d->product_id, 16); ack->product_id[16] = '\0';
	if ((t = strchr(ack->product_id, 0xFF))) *t = '\0';
	log_info("product ID is %s", ack->product_id);
	/* username */
	memcpy(ack->username, d->username, 16); ack->username[16] = '\0';
	if ((t = strchr(ack->username, 0xFF))) *t = '\0';
	log_info("username is %s", ack->username);
}

/**
 *	p7_recv_getcmd:
 *	Get data from command data field.
 *
 *	Layout is described in the fxReverse projet documentation.
 *
 *	@arg	cmd			the command packet destination
 *	@arg	data		the raw data
 */

static void p7_recv_getcmd(p7_packetcommand_t *cmd, void *data)
{
	char buf[9] = {0};
	char *d = data;

	/* get overwrite */
	memcpy(buf, data, 2);
	cmd->overwrite = (unsigned int)strtoul(buf, NULL, 16);
	/* - log - */
	log_info("overwrite mode is '%s' (%#02x)",
		p7_getowstring(cmd->overwrite), cmd->overwrite);

	/* get data type (?) */
	memcpy(buf, &d[2], 2);
	cmd->datatype = (int)strtoul(buf, NULL, 16);
	log_info("datatype is *todo: put MCS datatypes* (%#02x)", cmd->datatype);

	/* filesystem */
	memcpy(buf, &d[4], 8); buf[8] = '\0';
	cmd->filesize = (unsigned int)strtoul(buf, NULL, 16);
	log_info("filesize is %do", cmd->filesize);

	/* args length */
	buf[2] = '\0';
	memcpy(buf, &d[12], 2);
	size_t l1 = strtoul(buf, NULL, 16);
	memcpy(buf, &d[14], 2);
	size_t l2 = strtoul(buf, NULL, 16);
	memcpy(buf, &d[16], 2);
	size_t l3 = strtoul(buf, NULL, 16);
	memcpy(buf, &d[18], 2);
	size_t l4 = strtoul(buf, NULL, 16);
	memcpy(buf, &d[20], 2);
	size_t l5 = strtoul(buf, NULL, 16);
	memcpy(buf, &d[22], 2);
	size_t l6 = strtoul(buf, NULL, 16);

	/* args */
	char *p = &d[24];
	memcpy(cmd->args[0], p, l1); cmd->args[0][l1] = '\0'; p += l1;
	memcpy(cmd->args[1], p, l2); cmd->args[0][l2] = '\0'; p += l2;
	memcpy(cmd->args[2], p, l3); cmd->args[0][l3] = '\0'; p += l3;
	memcpy(cmd->args[3], p, l4); cmd->args[0][l4] = '\0'; p += l4;
	memcpy(cmd->args[4], p, l5); cmd->args[0][l5] = '\0'; p += l5;
	memcpy(cmd->args[5], p, l6); cmd->args[0][l6] = '\0';

	/* log args */
	if (l1) log_info("D1 is '%s'", cmd->args[0]);
	if (l2) log_info("D2 is '%s'", cmd->args[1]);
	if (l3) log_info("D3 is '%s'", cmd->args[2]);
	if (l4) log_info("D4 is '%s'", cmd->args[3]);
	if (l5) log_info("D5 is '%s'", cmd->args[4]);
	if (l6) log_info("D6 is '%s'", cmd->args[5]);
}

/**
 *	p7_recv_getdata:
 *	Get data from data packet data field.
 *
 *	@arg	data		organized data with fields
 *	@arg	raw			raw data
 *	@arg	raw_size	raw data size
 */

static void p7_recv_getdata(p7_packetdata_t *data, void *raw,
	unsigned int raw_size)
{
	packetdata_data_t *d = raw;
	char buf[5];

	/* total number */
	buf[4] = '\0';
	memcpy(buf, d->total_number, 4);
	data->total = (unsigned int)strtoul(buf, NULL, 16);
	log_info("Total data packets : %u", data->total);

	/* current id */
	memcpy(buf, d->current_number, 4);
	data->id = (unsigned int)strtoul(buf, NULL, 16);
	log_info("Data packet ID : %u", data->id);

	/* data */
	data->size = raw_size - 8;
	memcpy(data->data, d->data, data->size);
	log_info("Decoded data (%uo) :", data->size);
	logm_info(data->data, data->size);
}

/* ************************************************************************** */
/*  Main function                                                             */
/* ************************************************************************** */
/**
 *	p7_recv_main:
 *	Receive a packet.
 *
 *	@arg	packet		the packet data to fill
 *	@return				if it worked
 */

static int p7_recv_main(p7_packet_t *packet)
{
	/* packet data */
	unsigned char data[MAX_RAWDFLD_SIZE];
	unsigned int data_size = 0;

	/* Here's the direct buffer we will give to libusb. */
	unsigned char raw[MAX_PACKET_SIZE], *r = raw;
	unsigned int packet_size = 6;

	/* Here's the buffer we will use to convert to integer */
	char buf[5];

	/* Receive packet */
	log_info("receiving the packet...");
	int tf;
	while (1) {
		int err = libusb_bulk_transfer(p7_lusb_handler, ENDPOINT_IN, raw,
			MAX_PACKET_SIZE, &tf, p7_active ? ACTIVE_TIMEOUT : PASSVE_TIMEOUT);

		/* check error */
		switch (err) {
		/* - timeout - */
		case LIBUSB_ERROR_TIMEOUT:
			log_error("timeout received");
			p7_error = p7_error_timeout;
			return (0);

		/* - device error - */
		case LIBUSB_ERROR_PIPE:
		case LIBUSB_ERROR_NO_DEVICE:
			log_fatal("calculator was disconnected");
			p7_error = p7_error_nocalc;
			return (0);

		/* - unknown error - */
		case 0: break;
		default:
			log_fatal("libusb_bulk_transfer returned %d: %s", err,
				libusb_error_name(err));
			log_fatal("errno was %d: %s", errno, strerror(errno));
			p7_error = p7_error_unknown;
			return (0);
		}

		/* check empty packet */
		if (tf) break;
	}

	/* log packet */
	log_info("received the following packet :");
	logm_info(raw, (size_t)tf);

	/* Then get data from it */
	/* - type - */
	packet->type = *r++;

	/* - subtype - */
	buf[0] = (char)*r++, buf[1] = (char)*r++, buf[2] = '\0';
	unsigned int subtype = (unsigned int)strtoul(buf, NULL, 16);

	/* - extended - */
	if (*r++ == '1') {
		/* - get data size - */
		memcpy(buf, r, 4); buf[4] = '\0', r += 4;
		data_size = (unsigned int)strtoul(buf, NULL, 16);
		packet_size += 4 + data_size;

		/* - save current raw pointer - */
		unsigned char *rs = r;

		/* - move current raw pointer - */
		/* we want to do this as packet.size might change with data decoding. */
		r += data_size;

		/* - decode data - */
		data_size = p7_decode(data, rs, data_size);
	}

	/* get fields out for specific packets */
	switch (packet->type) {
	/* - for command - */
	case p7_pt_cmd:
		log_info("packet was interpreted as a command one");
		packet->command.code = subtype;
		log_info("command is '%s' (%#02x)",
			p7_getcmdstring(subtype), subtype);
		if (data_size) p7_recv_getcmd(&packet->command, data);
		break;

	/* - for data - */
	case p7_pt_data:
		log_info("packet was interpreted as a data one");
		p7_recv_getdata(&packet->data, data, data_size);
		break;

	/* - for roleswap - */
	case p7_pt_roleswp:
		log_info("packet was interpreted as a roleswap one");
		log_info("becoming active again");
		p7_active = 1;
		break;

	/* - for check - */
	case p7_pt_check:
		log_info("packet was interpreted as a check one");
		packet->check.initial = !subtype;
		break;

	/* - for ack - */
	case p7_pt_ack:
		log_info("packet was interpreted as an ack one");
		packet->ack.overwrite = (subtype == 0x01);
		packet->ack.extended = (subtype == 0x02);
		if (packet->ack.extended) p7_recv_getack(&packet->ack, data);
		break;

	/* - for error - */
	case p7_pt_error:
		log_info("packet was interpreted as an error one");
		log_info("error is '%s' (%#02x)",
			p7_geterrstring(subtype), subtype);
		packet->error.code = subtype;
		break;

	/* - for termination - */
	case p7_pt_terminate:
		log_info("packet was interpreted as a terminate one");
		log_info("termination cause is '%s' (%#02x)",
			p7_gettermstring(subtype), subtype);
		packet->term.reason = subtype;
		break;
	}

	/* and check the sum (yo) */
	memcpy(buf, r, 2); buf[2] = '\0', r += 2;
	if (p7_checksum(raw, packet_size) != (unsigned int)strtoul(buf, NULL, 16)) {
		p7_error = p7_error_checksum;
		return (0);
	}

	/* finally, return the packet */
	return (1);
}

/**
 *	p7_recv:
 *	Receives packet, checks for errors.
 *
 *	@arg	packet		the packet to fill out
 *	@return				if it worked
 */

int p7_recv(p7_packet_t *packet)
{
	int tries = 2, fail = 0;

	do {
		/* get packet */
		fail = !p7_recv_main(packet);

		/* if there was a fail */
		if (fail) {
			switch (p7_error) {
			/* if calc couldn't be found, terminate communication */
			case p7_error_nocalc:
				tries = 0;
				break;

			/* if it is a timeout, remove a try */
			case p7_error_timeout:
				tries--;
				break;

			/* if it is a bad checksum, ask for resend ; also, */
			/* reset tries number in case of bad checksum between timeouts */
			case p7_error_checksum:
				tries = 2;
				break;
			}
		}
	} while (fail && tries);

	/* if timeouts is the cause of failure, */
	/* terminate communication with packets */
	if (fail && p7_error == p7_error_timeout)
		p7_term();

	return (!fail);
}
