/* ************************************************************************** */
/*                                       _____           _                    */
/*  packet/send.c                       |_   _|__  _   _| |__   ___ _   _     */
/*  | Project : libp7                     | |/ _ \| | | | '_ \ / _ \ | | |    */
/*                                        | | (_) | |_| | | | |  __/ |_| |    */
/*  By: thomas <thomas@touhey.fr>         |_|\___/ \__,_|_| |_|\___|\__, |.fr */
/*  Last updated: 2016/08/16 12:43:57                               |___/     */
/*                                                                            */
/* ************************************************************************** */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <libp7/internals.h>

/**
 *	p7_response:
 *	Response buffer.
 */

p7_packet_t p7_response;

/**
 *	p7_response_got:
 *	Got response.
 */

int p7_response_got = 0;

/**
 *	p7_send_basic:
 *	Send a non-extended packet.
 *
 *	@arg	type		the packet type
 *	@arg	subtype		the packet subtype
 *	@return				if worked (0 if none)
 */

int p7_send_basic(unsigned int type, unsigned int subtype)
{
	unsigned char packet[6];

	/* prepare packet */
	packet[0] = (unsigned char)type;
	sprintf((char*)&packet[1], "%02X" "%0X", subtype, 0);
	sprintf((char*)&packet[4], "%02X", p7_checksum(packet, 6));

	/* log packet */
	log_info("sending the following basic packet :");
	logm_info(packet, 6);

	/* check if handler is initialized */
	if (!p7_lusb_handler) {
		p7_error = p7_error_uninitialized;
		return (0);
	}

	/* sending loop */
	int wasresend = 0;
	do {
		/* log resend */
		if (wasresend) log_warn("resend request was received, resend it goes");

		/* send prepared packet */
		int tf;
		int err = libusb_bulk_transfer(p7_lusb_handler, ENDPOINT_OUT,
			packet, 6, &tf, SENDNG_TIMEOUT);
		switch (err) {
		/* - timeout error - */
		case LIBUSB_ERROR_TIMEOUT:
			log_error("timeout received");
			p7_error = p7_error_timeout;
			return (0);

		/* - device disconnected - */
		case LIBUSB_ERROR_PIPE:
		case LIBUSB_ERROR_NO_DEVICE:
			log_fatal("calculator was disconnected");
			p7_error = p7_error_nocalc;
			return (0);

		/* - unknown error - */
		case 0: break;
		default:
			log_fatal("libusb_bulk_transfer returned %d: %s", err,
				libusb_error_name(err));
			log_fatal("errno was %d: %s", errno, strerror(errno));
			p7_error = p7_error_unknown;
			return (0);
		}

		/* set wasresend for logging */
		wasresend = 1;
	} while ((p7_response_got = p7_recv(&p7_response))
		&& p7_response.type == p7_pt_error
		&& p7_response.error.code == p7_ec_resend);

	/* packet sending is finished */
	return (p7_response_got);
}

/**
 *	p7_send_ext:
 *	Send an extended packet.
 *
 *	@arg	type		the packet type
 *	@arg	subtype		the packet subtype
 *	@arg	data		the packet data
 *	@arg	size		the packet data size
 */

int p7_send_ext(unsigned int type, unsigned int subtype,
	void *data, unsigned int size)
{
	/* create packet buffer */
	unsigned char packet[MAX_PACKET_SIZE];

	/* fill packet buffer */
	/* - first infos - */
	packet[0] = (unsigned char)type;
	sprintf((char*)&packet[1], "%02X" "%0X", subtype, 1);
	/* - data -
	 *	subtility: sprintf adds an '\0' at the end, so as we write after what
	 *		is before, we have to save the character that was there before
	 *		to restore it after.
	 */
	size = p7_encode(&packet[8], data, size);
	unsigned char c = packet[8];
	sprintf((char*)&packet[4], "%04X", size);
	packet[8] = c; /* as sprintf adds an '\0' */
	/* - checksum - */
	sprintf((char*)&packet[8 + size], "%02X",
		p7_checksum(packet, 8 + size + 2));

	/* check if handler is initialized */
	if (!p7_lusb_handler) {
		p7_error = p7_error_uninitialized;
		return (0);
	}

	/* log packet */
	log_info("sending the following extended packet :");
	logm_info(packet, 8 + size + 2);

	/* sending loop */
	int wasresend = 0;
	p7_error = p7_error_timeout;
	do {
		/* log resend request */
		if (wasresend) log_warn("resend request was received, resend it goes");
	
		/* send prepared packet */
		int transferred;
		int err = libusb_bulk_transfer(p7_lusb_handler, ENDPOINT_OUT,
			packet, 8 + (int)size + 2, &transferred, SENDNG_TIMEOUT);
		switch (err) {
		/* - timeout error - */
		case LIBUSB_ERROR_TIMEOUT:
			log_error("timeout received");
			p7_error = p7_error_timeout;
			return (0);

		/* - device disconnected - */
		case LIBUSB_ERROR_PIPE:
		case LIBUSB_ERROR_NO_DEVICE:
			log_fatal("calculator was disconnected");
			p7_error = p7_error_nocalc;
			return (0);

		/* - unknown error - */
		case 0: break;
		default:
			log_fatal("libusb_bulk_transfer returned %d: %s", err,
				libusb_error_name(err));
			log_fatal("errno was %d: %s", errno, strerror(errno));
			p7_error = p7_error_unknown;
			return (0);
		}

		/* set wasreset for logging */
		wasresend = 1;
	} while ((p7_response_got = p7_recv(&p7_response))
		&& p7_response.type == p7_pt_error
		&& p7_response.error.code == p7_ec_resend);

	/* packet sending is finished */
	return (p7_response_got);
}
