/* ************************************************************************** */
/*                                       _____           _                    */
/*  protocol/reqfile.c                  |_   _|__  _   _| |__   ___ _   _     */
/*  | Project : libp7                     | |/ _ \| | | | '_ \ / _ \ | | |    */
/*                                        | | (_) | |_| | | | |  __/ |_| |    */
/*  By: thomas <thomas@touhey.fr>         |_|\___/ \__,_|_| |_|\___|\__, |.fr */
/*  Last updated: 2016/08/20 22:14:42                               |___/     */
/*                                                                            */
/* ************************************************************************** */

#include <string.h>
#include <libp7/internals.h>

/**
 *	p7_reqfile:
 *	Requests a file.
 *
 *	@arg	file		the file (should be writable)
 *	@arg	dirname		the directory name
 *	@arg	filename	the filename
 *	@arg	devname		the device name
 *	@return				if it worked
 */

int p7_reqfile(FILE *file, char *dirname, char *filename, char *devname)
{
	p7_packet_t packet;

	/* check filename */
	if (strnlen(filename, 13) > 12) {
		p7_error = p7_error_filename;
		log_error("sent filename was crappy, returning error to user");
		return (0);
	}

	/* check dirname */
	if (dirname && strnlen(dirname, 9) > 8) {
		p7_error = p7_error_dirname;
		log_error("sent dirname was crappy, returning error to user");
		return (0);
	}

	/* send command packet */
	log_info("sending file transfer request packet");
	if (!p7_send_cmdfls_reqfile(dirname, filename, devname)) {
		log_fatal("couldn't send file transfer request packet");
		return (0);
	}

	/* get response packet */
	log_info("receiving file transfer request packet response");
	if (!p7_receive(&packet)) {
		log_fatal("didn't receive file transfer request answer");
		return (0);
	}

	/* check it */
	if (packet.type == p7_pt_error
	 && packet.error.code == p7_ec_overwrite_impossible) {
		p7_error = p7_error_overwrite_impossible;
		log_fatal("wrong device name or requested file doesn't exist");
		return (0);
	} else if (packet.type != p7_pt_ack) {
		p7_error = p7_error_unknown;
		log_fatal("didn't receive ack or error...");
		return (0);
	}

	/* swap roles */
	log_info("sending roleswap");
	if (!p7_send_roleswp()) {
		log_fatal("couldn't swap roles");
		return (0);
	}

	/* receive response file transfer command */
	log_info("receiving command with filesize");
	if (!p7_receive(&packet) || packet.type != p7_pt_cmd) {
		log_fatal("didn't receive expected command");
		return (0);
	}
	/* -- get filesize from packet ? -- */

	/* - Note: we are now in passive mode - */
	/* responding with acks */
	log_info("starting main loop");
	while (1) {
		/* send ack */
		log_info("send ack");
		if (!p7_send_ack()) {
			log_fatal("couldn't send ack");
			return (0);
		}

		/* get packet */
		log_info("receive packet");
		if (!p7_receive(&packet)) {
			log_fatal("could not get packet");
			return (0);
		}

		/* check what to do with packet */
		if (packet.type == p7_pt_data) {
			log_info("is data, send it to filestream");
			fwrite(packet.data.data, 1, packet.data.size, file);
		} else if (packet.type == p7_pt_roleswp) {
			log_info("is roleswap, our role as passive is now done");
			break;
		}
	}

	/* file request has been done */
	return (1);
}
