/* ************************************************************************** */
/*                                       _____           _                    */
/*  protocol/sendfile.c                 |_   _|__  _   _| |__   ___ _   _     */
/*  | Project : libp7                     | |/ _ \| | | | '_ \ / _ \ | | |    */
/*                                        | | (_) | |_| | | | |  __/ |_| |    */
/*  By: thomas <thomas@touhey.fr>         |_|\___/ \__,_|_| |_|\___|\__, |.fr */
/*  Last updated: 2016/08/19 14:55:33                               |___/     */
/*                                                                            */
/* ************************************************************************** */

#include <string.h>
#include <libp7/internals.h>

/**
 *	p7_sendstream:
 *	Sends a stream.
 *
 *	@arg	file		the file stream
 *	@arg	size		the file size
 *	@arg	dirname		the directory name
 *	@arg	filename	the file name
 *	@arg	devname		the device name
 *	@arg	overwrite	if no confirmation required, should overwrite or not
 *	@arg	ow_conf		if file exists and this callback is not NULL,
 *						call it. If answer is 1, overwrite, otherwise, not.
 *	@return				if it worked
 */

int p7_sendstream(FILE *file, unsigned int size, char *dirname, char *filename,
	char *devname, int overwrite, int (*ow_conf)(void))
{
	p7_packet_t packet;

	/* check filename */
	if (strnlen(filename, 13) > 12) {
		p7_error = p7_error_filename;
		log_error("sent filename was crappy, returning error to user");
		return (0);
	}

	/* check dirname */
	if (dirname && strnlen(dirname, 9) > 8) {
		p7_error = p7_error_dirname;
		log_error("sent dirname was crappy, returning error to user");
		return (0);
	}

	/* check if fits in 256Kio -- so that it sticks for sure in 32-bits */
	if (size >= 262144) {
		p7_error = p7_error_fullmem;
		log_error("sent file is too large, returning error to user");
		return (0);
	}

	/* choose overwrite mode
	 *	If we want to never overwrite, we don't want the calculator to terminate
	 *	communication, so we ask confirmation and we will systematically
	 *	return no.
	 */
	p7_commandoverwrite_t owmode;
	if (ow_conf || !overwrite)
		owmode = p7_cow_request_confirmation_before_overwriting;
	else owmode = p7_cow_force_overwrite;

	/* send command packet */
	log_info("sending file transfer packet");
	if (!p7_send_cmdfls_sendfile(owmode, size, dirname, filename, devname)) {
		log_fatal("couldn't send file transfer packet");
		return (0);
	}

	/* get response packet */
	log_info("receiving file transfer answer");
	if (!p7_receive(&packet)
	 || (packet.type != p7_pt_ack && packet.type != p7_pt_error)) {
		log_fatal("didn't receive file transfer answer");
		return (0);
	}
	if (packet.type == p7_pt_error) {
		switch (packet.error.code) {
		/* file exists, confirm or not */
		case p7_ec_overwrite:
			log_info("calc wanted overwrite confirmation");
			/* check whether to ask */
			int denied = 0;
			if (ow_conf && (*ow_conf)()) {
				log_info("sending yes to overwrite confirmation");
				if (!p7_send_ack_confirm_ow()) return (0);
			} else {
				log_info("sending no to overwrite confirmation");
				if (!p7_send_err_overwrite_no()) return (0);
				denied = 1;
				p7_error = p7_error_denied_overwrite;
			}
			/* get ack packet */
			if (!p7_receive(&packet) || packet.type != p7_pt_ack) {
				p7_error = p7_error_unknown;
				return (0);
			}
			/* if we denied, stop here
			 *	not really an error, but this is to say to the user that
			 *	file transfer did not happen.
			 *
			 *	we used the "denied" var instead of p7_error directly as
			 *	p7_error might be uninitialised and contain
			 *	p7_error_denied_overwrite from a previous call to sendfile.
			 */
			if (denied) {
				p7_error = p7_error_denied_overwrite;
				return (0);
			}
			break;

		/* impossible overwrite */
		case p7_ec_overwrite_noresponse:
		case p7_ec_overwrite_impossible:
			p7_error = p7_error_overwrite_impossible;
			log_error("overwrite impossible");
			return (0);

		/* memory full */
		case p7_ec_fullmem:
			p7_error = p7_error_fullmem;
			log_error("distant memory is full");
			return (0);

		/* resend error - just here to stfu the warning */
		case p7_ec_resend:
			p7_error = p7_error_unknown;
			return (0);
		}
	}

	/* send data */
	unsigned int lastpacket_size = size % MAX_RAWDATA_SIZE;
	unsigned int total = size / MAX_RAWDATA_SIZE
		+ !!lastpacket_size; /* left data */
	unsigned char buf[MAX_RAWDATA_SIZE];

	log_info("about to send %u packets", total);
	if (lastpacket_size)
		log_info("last one is irregular and contains %uo", lastpacket_size);
	for (unsigned int id = 1; id <= total; id++) {
		/* get datasize */
		unsigned int datasize = (id == total && lastpacket_size) ?
			lastpacket_size : MAX_RAWDATA_SIZE;

		/* fill buffer */
		fread(buf, sizeof(char), datasize, file);

		/* logging */
		log_info("sending packet %u/%u (%uo)", id, total, datasize);

		/* send data packet */
		p7_send_data(total, id, buf, datasize);

		/* check response */
		if (!p7_receive(&packet) || packet.type != p7_pt_ack)
			return (0);
	}

	/* everything went alright */
	return (1);
}

/**
 *	p7_sendfile:
 *	Sends a file.
 *
 *	@arg	file		the file stream (must be seekable !)
 *	@arg	dirname		the directory name
 *	@arg	filename	the file name
 *	@arg	devname		the device name
 *	@arg	overwrite	if no confirmation required, should overwrite or not
 *	@arg	ow_conf		if file exists and this callback is not NULL,
 *						call it. If answer is 1, overwrite, otherwise, not.
 *	@return				if it worked
 */

int p7_sendfile(FILE *file, char *dirname, char *filename,
	char *devname, int overwrite, int (*ow_conf)(void))
{
	/* calculate file size */
	log_info("calculating given file size");
	unsigned int size;
	if (fseek(file, 0, SEEK_END)) {
		p7_error = p7_error_noseek;
		log_error("could not seek in file");
		return (0);
	}
	size = (unsigned int)ftell(file);
	rewind(file);
	/* - log - */
	log_info("filesize is %uo", size);

	/* send file */
	return (p7_sendstream(file, size, dirname, filename, devname,
		overwrite, ow_conf));
}
